#include "stdlib.h"
#include "ctype.h"
#include "errno.h"
#include <limits.h>

unsigned long
strtoul (const char *cp, char **endp, int base)
{
  unsigned long result = 0,value;
  unsigned long maxres, maxval;
  int negative = 0;

#ifdef isspace
  while (isspace (*cp))
    cp++;
#endif

  if (cp[0] == '+' && isdigit (cp[1]))
    cp++;
  else if (cp[0] == '-' && isdigit (cp[1]))
    {
      negative = 1;
      cp++;
    }

  if (!base)
    {
      base = 10;
      if (*cp == '0')
	{
	  base = 8;
	  cp++;
	  if ((*cp == 'x') && isxdigit(cp[1]))
	    {
	      cp++;
	      base = 16;
	    }
	}
    }

  maxres = ULONG_MAX / base;
  maxval = ULONG_MAX % base;

  while (1)
    {
      value = *cp - '0';

      if (value < 10)
	;
      else if (value < 'A' - '0')
	break;
      else if (value <= 'Z' - '0')
	value -= 'A' - '0' + 10;
      else if (value < 'a' - '0')
	break;
      else if (value <= 'z' - '0')
	value -= 'a' - '0' + 10;
      else
	break;
      if (value >= base)
	break;
      if (result > maxres || (result == maxres && value > maxval))
	{
	  errno = ERANGE;
	  result = ULONG_MAX;
	  negative = 0;
	  break;
	}

      result = result * base + value;
      cp++;
    }

  if (endp)
    *endp = (char *) cp;
  return negative ? -result : result;
}
